using System;
using System.Reflection;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using VersionTask;

namespace UnitTests
{
    [TestClass]
    public class VersionInfoTests
    {

        #region Tests
        
        [TestMethod]
        public void VersionInfo_SameVersion_Equal()
        {
            var version1 = CreateVersion(1, 1, 1);
            var version2 = CreateVersion(1, 1, 1);

            var same = CallOperatorReturnBool(Operators.Equality, version1, version2);
            
            Assert.IsTrue(same);
        }
        
        [TestMethod]
        public void VersionInfo_DifferentPatch_NotEqual()
        {
            var version1 = CreateVersion(1, 1, 1);
            var version2 = CreateVersion(1, 1, 2);

            var same = CallOperatorReturnBool(Operators.Equality, version1, version2);

            Assert.IsFalse(same);
        }       
        
        [TestMethod]
        public void VersionInfo_FirstGreaterMajorSecondLowerMinor_GreaterFirst()
        {
            var version1 = CreateVersion(2, 1, 1);
            var version2 = CreateVersion(1, 1, 30);

            var same = CallOperatorReturnBool(Operators.GreaterThan, version1, version2);
            
            Assert.IsTrue(same);
        }        
        
        [TestMethod]
        public void VersionInfo_FirstGreaterMinorSecondGreaterMajorDifferentPatches_GreaterSecond()
        {
            var version1 = CreateVersion(1, 2, 1);
            var version2 = CreateVersion(2, 1, 30);

            var same = CallOperatorReturnBool(Operators.GreaterThan, version1, version2);
            
            Assert.IsFalse(same);
        }   
        
        [TestMethod]
        public void VersionInfo_MinorDifferent_NotEqual()
        {
            var version1 = CreateVersion(1, 2, 1);
            var version2 = CreateVersion(1, 1, 1);

            var same = CallOperatorReturnBool(Operators.Equality, version1, version2);
            
            Assert.IsFalse(same);
        }
        
        [TestMethod]
        public void VersionInfo_Addition_ReturnRightValue()
        {
            var version1 = CreateVersion(1, 2, 1);
            var version2 = CreateVersion(1, 1, 1);

            var expected = CreateVersion(2, 3, 2);

            var actual = CallOperatorReturnIVersion(Operators.Addition, version1, version2);
            var same = CallOperatorReturnBool(Operators.Equality, actual, expected);
            
            Assert.IsTrue(same);
        }
        
        
        [TestMethod]
        public void VersionInfo_Subtraction_ReturnRightValue()
        {
            var version1 = CreateVersion(3, 2, 2);
            var version2 = CreateVersion(1, 1, 1);

            var expected = CreateVersion(2, 1, 1);

            var actual = CallOperatorReturnIVersion(Operators.Subtraction, version1, version2);
            var same = CallOperatorReturnBool(Operators.Equality, actual, expected);
            
            Assert.IsTrue(same);
        }
        
        #endregion

        #region Private methods
        
        private VersionInfo CreateVersion(ulong major, ulong minor, ulong patch)
        {
            var instance = (VersionInfo)Activator.CreateInstance(Settings.VersionInfoType, major, minor, patch);
            return instance;
        }

        private bool CallOperatorReturnBool(Operators operatorCall, VersionInfo version1, VersionInfo version2)
        {
            var mi = version1.GetType().GetMethod($"op_{operatorCall}", BindingFlags.Static | BindingFlags.Public);
            var res = mi.Invoke(null, new[] { version1, version2 });
            return (bool) res;
        }
        
        
        private VersionInfo CallOperatorReturnIVersion(Operators operatorCall, VersionInfo version1, VersionInfo version2)
        {
            var mi = version1.GetType().GetMethod($"op_{operatorCall}", BindingFlags.Static | BindingFlags.Public);
            var res = mi.Invoke(null, new[] { version1, version2 });
            return (VersionInfo) res;
        }

        #endregion
    }
}